"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = createGenerator;

var _adapters = _interopRequireDefault(require("./adapters"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance"); }

function _iterableToArray(iter) { if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var validPseudoClasses = ['link', 'hover', 'focus', 'active', 'visited', 'disabled', 'focusWithin', 'firstChild', 'lastChild'];
var validMediaQueries = ['viewportWidth', 'viewportHeight'];

function createGenerator() {
  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$adapter = _ref.adapter,
      adapter = _ref$adapter === void 0 ? 'fela' : _ref$adapter;

  var usedAdapter = _adapters.default.find(function (adapt) {
    return adapt.name === adapter;
  });

  return function generate(ast, fileName) {
    var css = generateCSS(ast);
    var js = generateJS(ast, usedAdapter);
    var root = generateRoot(ast);
    return _objectSpread({
      _root: root
    }, css, js);
  };
}

function generateRoot(ast) {
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  var imports = styles.map(function (module) {
    return 'import { ' + module.name + " } from './" + module.name + ".elodin.js'";
  }).join('\n');
  return imports + '\n\n' + 'export {\n  ' + styles.map(function (module) {
    return module.name;
  }).join(',\n  ') + '\n}';
}

function generateCSSValue(value) {
  var unit = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

  if (value.type === 'NumericLiteral') {
    return value.value + (unit ? 'px' : '');
  }

  if (value.type === 'FunctionExpression') {
    return value.callee + '(' + value.params.map(function (param) {
      return generateCSSValue(param, false);
    }).join(',') + ')';
  }

  if (value.type === 'Float') {
    return value.integer + '.' + value.fractional + (unit ? 'px' : '');
  }

  return value.value;
}

function generateCSS(ast) {
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  return styles.reduce(function (files, module) {
    var classes = generateClasses(module.body);
    files[module.name + '.elodin.css'] = classes.filter(function (selector) {
      return selector.declarations.length > 0;
    }).map(function (selector) {
      var css = stringifyCSS(selector.declarations, module.name + selector.pseudo);

      if (selector.media) {
        return '@media ' + selector.media + ' {\n' + css + '\n}';
      }

      return css;
    }).join('\n\n');
    return files;
  }, {});
}

function generateClasses(nodes) {
  var classes = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  var pseudo = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '';
  var media = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : '';
  var base = nodes.filter(function (node) {
    return node.type === 'Declaration';
  });
  var nesting = nodes.filter(function (node) {
    return node.type !== 'Declaration';
  });
  classes.push({
    media: media,
    pseudo: pseudo,
    declarations: getStaticDeclarations(base)
  });
  nesting.forEach(function (nest) {
    if (nest.property.type === 'Variable' && nest.property.environment) {
      if (nest.boolean && validPseudoClasses.indexOf(nest.property.value) !== -1) {
        generateClasses(nest.body, classes, pseudo + ':' + nest.property.value, media);
      }

      if (validMediaQueries.indexOf(nest.property.value) !== -1) {
        generateClasses(nest.body, classes, pseudo, getMediaQuery(nest.value.value, nest.property.value, nest.operator));
      }
    }
  });
  return classes;
}

function getStaticDeclarations(declarations) {
  return declarations.filter(function (decl) {
    return decl.value.type !== 'Variable';
  }).map(function (declaration) {
    return {
      property: declaration.property,
      value: generateCSSValue(declaration.value)
    };
  });
}

var uppercasePattern = /[A-Z]/g;
var msPattern = /^ms-/;
var cache = {};

function hyphenateProperty(string) {
  return string in cache ? cache[string] : cache[string] = string.replace(uppercasePattern, '-$&').toLowerCase().replace(msPattern, '-ms-');
}

function stringifyCSS(declarations, name) {
  return '.' + name + ' {\n  ' + declarations.map(function (decl) {
    return hyphenateProperty(decl.property) + ': ' + decl.value;
  }).join(';\n  ') + '\n' + '}';
}

function generateJS(ast, adapter) {
  // TODO: include fragments
  var styles = ast.body.filter(function (node) {
    return node.type === 'Style';
  });
  return styles.reduce(function (files, module) {
    var style = generateStyle(module.body);
    files[module.name + '.elodin.js'] = adapter.stringify(style, module.name);
    return files;
  }, {});
}

function generateStyle(nodes) {
  var base = nodes.filter(function (node) {
    return node.type === 'Declaration';
  });
  var nestings = nodes.filter(function (node) {
    return node.type !== 'Declaration';
  });
  var declarations = base.filter(function (decl) {
    return decl.value.type === 'Variable';
  }).map(function (declaration) {
    return {
      property: declaration.property,
      value: declaration.value.value
    };
  });
  var nests = nestings.map(function (nest) {
    if (nest.property.type === 'Variable' && nest.property.environment) {
      if (nest.boolean && validPseudoClasses.indexOf(nest.property.value) !== -1) {
        return {
          property: ':' + nest.property.value,
          value: generateStyle(nest.body)
        };
      }

      if (validMediaQueries.indexOf(nest.property.value) !== -1) {
        return {
          property: getMediaQuery(nest.value.value, nest.property.value, nest.operator),
          value: generateStyle(nest.body)
        };
      }
    }
  }).filter(function (nesting) {
    return nesting && nesting.value.length > 0;
  });
  return [].concat(_toConsumableArray(declarations), _toConsumableArray(nests));
}

function getMediaQuery(value, property, operator) {
  var dimension = property.indexOf('Height') !== -1 ? 'height' : 'width';

  if (operator === '=') {
    return '(min-' + dimension + ': ' + value + 'px) and (max-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '>') {
    return '(min-' + dimension + ': ' + (value + 1) + 'px)';
  }

  if (operator === '>=') {
    return '(min-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '<=') {
    return '(max-' + dimension + ': ' + value + 'px)';
  }

  if (operator === '<') {
    return '(max-' + dimension + ': ' + (value - 1) + 'px)';
  }
}